@@ -1,16 +1,10 @@
 import math
-import itertools
-import sys
-IS_PYTHON3 = sys.version_info[0] >= 3
+from io import StringIO
 
-if IS_PYTHON3:
-    exec('from ._edit_descriptors import *')
-    exec('from ._misc import expand_edit_descriptors, has_next_iterator')
-    exec('from . import config')
-else:
-    exec('from _edit_descriptors import *')
-    exec('from _misc import expand_edit_descriptors, has_next_iterator')
-    exec('import fortranformat.config as config')
+from ._edit_descriptors import *
+from ._exceptions import InvalidFormat
+from ._misc import expand_edit_descriptors, has_next_iterator
+from . import config
 
 
 PROC_SIGN_ZERO = config.PROC_SIGN_ZERO
@@ -19,18 +13,19 @@ PROC_DECIMAL_CHAR = config.PROC_DECIMAL_CHAR
 G0_NO_BLANKS = config.G0_NO_BLANKS
 PROC_NO_LEADING_BLANK = config.PROC_NO_LEADING_BLANK
 
+
 def output(eds, reversion_eds, values):
     '''
     a function to take a list of valid f77 edit descriptors and respective values
     and output the corresponding string
     '''
-    record = ''
+    record = StringIO()
     state = {
-        'position' : 0,
-        'scale' : 0,
-        'incl_plus' : False,
-        'blanks_as_zeros' : False,
-        'halt_if_no_vals' : False,
+        'position': 0,
+        'scale': 0,
+        'incl_plus': False,
+        'blanks_as_zeros': False,
+        'halt_if_no_vals': False,
     }
 
     # if format is empty with no values specified, then output blank record -
@@ -39,7 +34,7 @@ def output(eds, reversion_eds, values):
     # check that if there is a reversion, that values can be output
     reversion_contains_output_ed = False
     for ed in reversion_eds:
-        if isinstance(ed, OUTPUT_EDS):
+        if ed.is_output:
             reversion_contains_output_ed = True
             break
     # Get full list of edit descriptors
@@ -49,126 +44,108 @@ def output(eds, reversion_eds, values):
     get_ed = has_next_iterator(eds)
     get_value = has_next_iterator(values)
     tmp_reversion_eds = []
+    is_complete = False
     # continue until out of edit descriptors or values
     while True:
-        # no more edit descriptors, no more values, stop output, 
-        # todo: this will cut short a reversion edit descriptor section - is this right?
-        if not get_ed.has_next() and not get_value.has_next():
+        # no more edit descriptors, no more values, stop output
+        if not get_ed.has_next() \
+                and not get_value.has_next() \
+                and (not reversion_contains_output_ed or not len(tmp_reversion_eds)):
             break
         # take a edit descriptor off the queue if there is any
         if get_ed.has_next():
-            if IS_PYTHON3:
-                ed = next(get_ed)
-            else:
-                ed = get_ed.next()
+            ed = next(get_ed)
         else:
-            if reversion_contains_output_ed == True:
+            if reversion_contains_output_ed:
                 # take from reversion edit descriptors if there is a value
                 # requiring output still
+
                 while True:
                     if len(tmp_reversion_eds):
                         ed = tmp_reversion_eds.pop()
                         # these edit descriptors are ignored in reversion state
-                        if not isinstance(ed, NON_REVERSION_EDS):
+                        if not ed.is_non_reversion:
                             break
                     else:
                         # Regardless of where cursor is, is moved to the
                         # next record
-                        record = record + config.RECORD_SEPARATOR
-                        state['position'] = len(record)
-                        tmp_reversion_eds = reversion_eds[::-1]
+                        record.write(config.RECORD_SEPARATOR)
+                        state['position'] = record.tell()
+                        if get_value.has_next():
+                            # Only reset the reversion if we have more values to output
+                            tmp_reversion_eds = reversion_eds[::-1]
+                        else:
+                            # ... otherwise we are done
+                            is_complete = True
+                            break
+                if is_complete:
+                    break
             else:
-                # ignore the revsion edit descriptors as cannot output the
+                # ignore the revsersion edit descriptors as cannot output the
                 # final value
                 break
         # check if edit descriptor requires a value
-        if isinstance(ed, OUTPUT_EDS):
+        if ed.is_output:
             if get_value.has_next():
-                if IS_PYTHON3:
-                    val = next(get_value)
-                else:
-                    val = get_value.next()
+                val = next(get_value)
             else:
                 # is a edit descriptor that requires a value but no value
-                # todo: does it stop gracefully or raise error?
-                break   
-            if isinstance(ed, I):
-                sub_string = _compose_i_string(ed.width, ed.min_digits, state, val)
-            elif isinstance(ed, B):
-                w = ed.width
-                m = ed.min_digits
-                sub_string = _compose_boz_string(w, m, state, val, 'B')
-            if isinstance(ed, O):
-                w = ed.width
-                m = ed.min_digits
-                sub_string = _compose_boz_string(w, m, state, val, 'O')
-            if isinstance(ed, Z):
-                w = ed.width
-                m = ed.min_digits
-                sub_string = _compose_boz_string(w, m, state, val, 'Z')
-            elif isinstance(ed, F):
-                w = ed.width
-                e = None
-                d = ed.decimal_places
-                sub_string = _compose_float_string(w, e, d, state, val, 'F')
-            elif isinstance(ed, E):
-                w = ed.width
-                e = ed.exponent
-                d = ed.decimal_places
-                sub_string = _compose_float_string(w, e, d, state, val, 'E')
-            elif isinstance(ed, D):
-                w = ed.width
-                e = None
-                d = ed.decimal_places
-                sub_string = _compose_float_string(w, e, d, state, val, 'D')
-            elif isinstance(ed, G):
-                w = ed.width
-                e = ed.exponent
-                d = ed.decimal_places
-                sub_string = _compose_float_string(w, e, d, state, val, 'G')
-            elif isinstance(ed, EN):
-                w = ed.width
-                e = ed.exponent
-                d = ed.decimal_places
-                sub_string = _compose_float_string(w, e, d, state, val, 'EN')
-            elif isinstance(ed, ES):
-                w = ed.width
-                e = ed.exponent
-                d = ed.decimal_places
-                sub_string = _compose_float_string(w, e, d, state, val, 'ES')
-            elif isinstance(ed, L):
+                # we simply ignore at this point - likely this is an incomplete reversion
+                break
+            if ed.name == "I":
+                if val is None:
+                    val = 0
+                sub_string = _compose_i_string(
+                    ed.width, ed.min_digits, state, val)
+            elif ed.name in ("B", "O", "Z"):
+                if val is None:
+                    val = 0
+                sub_string = _compose_boz_string(ed.width, ed.min_digits, state, val, ed.name)
+            elif ed.name in ("D", "E", "EN", "ES", "F", "G"):
+                if val is None:
+                    val = 0.0
+                sub_string = _compose_float_string(ed.width, ed.exponent, ed.decimal_places, state, val, ed.name)
+            elif ed.name == "L":
+                if val is None:
+                    val = True
                 sub_string = _compose_l_string(ed.width, state, val)
-            elif isinstance(ed, A):
+            elif ed.name == "A":
+                if val is None:
+                    val = ''
                 sub_string = _compose_a_string(ed.width, state, val)
-            state['position'], record = _write_string(record, sub_string, state['position'])
+            state['position'], record = _write_string(
+                record, sub_string, state['position'])
         else:
             # token does not require a value
-            if isinstance(ed, (S, SS)):
+            if ed.name in ("S", "SS"):
                 state['incl_plus'] = False
-            if isinstance(ed, SP):
+            if ed.name == "SP":
                 state['incl_plus'] = True
-            elif isinstance(ed, P):
+            elif ed.name == "P":
                 state['scale'] = ed.scale
-            elif isinstance(ed, BN):
+            elif ed.name == "BN":
                 # This is moot since for output, this does not do anything
                 state['blanks_as_zeros'] = False
-            elif isinstance(ed, BZ):
+            elif ed.name == "BZ":
                 state['blanks_as_zeros'] = True
-            elif isinstance(ed, Colon):
+            elif ed.name == "Colon":
                 state['halt_if_no_vals'] = True
-            elif isinstance(ed, Slash):
-                state['position'], record = _write_string(record, config.RECORD_SEPARATOR, state['position'])
-            elif isinstance(ed, (X, TR)):
+            elif ed.name == "Slash":
+                state['position'], record = _write_string(
+                    record, config.RECORD_SEPARATOR, state['position'])
+            elif ed.name in ("X", "TR"):
                 state['position'] = state['position'] + ed.num_chars
-            elif isinstance(ed, TL):
+            elif ed.name == "TL":
                 state['position'] = state['position'] - ed.num_chars
-            elif isinstance(ed, T):
+            elif ed.name == "T":
                 state['position'] = ed.num_chars - 1
-            elif isinstance(ed, QuotedString):
+            elif ed.name == "QuotedString":
                 sub_string = ed.char_string
-                state['position'], record = _write_string(record, sub_string, state['position'])
+                state['position'], record = _write_string(
+                    record, sub_string, state['position'])
     # output the final record
-    return record
+    return record.getvalue()
+
 
 def _compose_nan_string(w, ftype):
     if ftype in ['B', 'O', 'Z']:
@@ -176,11 +153,12 @@ def _compose_nan_string(w, ftype):
     else:
         # Allow at least 'NaN' to be printed
         if w == 0:
-            w = 4 # n.b. this is what is set in Gfortran 4.4.0
+            w = 4  # n.b. this is what is set in Gfortran 4.4.0
         if w < 3:
             return '*' * w
     return 'NaN'.rjust(w)
 
+
 def _compose_inf_string(w, ftype, sign_bit):
     if ftype in ['B', 'O', 'Z']:
         return ''
@@ -206,16 +184,22 @@ def _compose_inf_string(w, ftype, sign_bit):
         # Should only output short version with no sign if positive
         else:
             return 'Inf'
-                
+
 
 def _compose_float_string(w, e, d, state, val, ftype):
     '''
-    Adapted from code in glibfortran which is written in C so is somwhat
+    Adapted from code in glibfortran which is written in C so is somewhat
     'bit-pushy' in nature. Writes the value to an initial string (buffer)
     and then pulls the subsequent strings from that
     '''
     if (d < 0) or (d is None):
         raise InvalidFormat('Unspecified precision')
+
+    # It seems that 0 is not actually permitted as the number of decimal places for an E format
+    # This returns asterisks in the Intel compiler
+    if (d == 0) and (ftype == 'E'):
+        return '*' * w
+
     # Make sure they are ints
     d = int(round(d))
     if e is not None:
@@ -223,9 +207,9 @@ def _compose_float_string(w, e, d, state, val, ftype):
     if w is not None:
         w = int(round(w))
     # ==== write_float ==== (function)
-    edigits = 4 # Largest number of exponent digits expected
+    edigits = 4  # Largest number of exponent digits expected
     if (ftype in ['F', 'EN', 'G']) or \
-        ((ftype in ['D', 'E']) and (state['scale'] != 0)):
+            ((ftype in ['D', 'E']) and (state['scale'] != 0)):
         # Convert with full possible precision
         ndigits = PROC_MIN_FIELD_WIDTH - 4 - edigits
     else:
@@ -243,11 +227,11 @@ def _compose_float_string(w, e, d, state, val, ftype):
     else:
         sign_bit = val < 0
     # handle the nan and inf cases
-    if type(val) is float and val != val :
-        return _compose_nan_string(w, ed)
+    if type(val) is float and val != val:
+        return _compose_nan_string(w, ftype)
     Infinity = 1e1000000
     if val in (-Infinity, Infinity):
-        return _compose_inf_string(w, ed, sign_bit)
+        return _compose_inf_string(w, ftype, sign_bit)
     tmp = abs(val)
     # Round the input if the input is less than 1
     if (ftype == 'F') and (d == state['scale']) and (d == 0):
@@ -256,22 +240,17 @@ def _compose_float_string(w, e, d, state, val, ftype):
     zero_flag = (tmp == 0)
     # === DTOA === (macro)
     # write the tmp value to the string buffer
-    # sprintf seems to allow negative number of decimal places, need to correct for this
-    if ndigits <= 0:
-        fmt = '%+-#' + str(PROC_MIN_FIELD_WIDTH) + 'e'
-    else:
-        fmt = '%+-#' + str(PROC_MIN_FIELD_WIDTH) + '.' + str(ndigits - 1) + 'e'
-    buff = fmt % tmp
+    buff = f"{tmp:<+#{PROC_MIN_FIELD_WIDTH}.{max(0, ndigits - 1)}e}"
     # === WRITE_FLOAT === (macro)
     if ftype != 'G':
-        return  _output_float(w, d, e, state, ftype, buff, sign_bit, zero_flag, ndigits, edigits)
+        return _output_float(w, d, e, state, ftype, buff, sign_bit, zero_flag, ndigits, edigits)
     else:
         # Perform different actions for G edit descriptors depending on value
         #
         # Generate corresponding I/O format for FMT_G and output.
         # The rules to translate FMT_G to FMT_E or FMT_F from DEC fortran
         # LRM (table 11-2, Chapter 11, "I/O Formatting", P11-25) is:
-        # 
+        #
         # Data Magnitude                              Equivalent Conversion
         # 0< m < 0.1-0.5*10**(-d-1)                   Ew.d[Ee]
         # m = 0                                       F(w-n).(d-1), n' '
@@ -281,25 +260,33 @@ def _compose_float_string(w, e, d, state, val, ftype):
         # ................                           ..........
         # 10**(d-1)-0.5*10**(-1)<= m <10**d-0.5       F(w-n).0,n(' ')
         # m >= 10**d-0.5                              Ew.d[Ee]
-        # 
+        #
         # notes: for Gw.d ,  n' ' means 4 blanks
         #        for Gw.dEe, n' ' means e+2 blanks
+
+        # ^^^^ no longer where the above is from, possibly the source of GFortran originally and was included
+        # to illuminate the direct ported code below. However it's clear from other sources that a 'd' value
+        # of zero is not valid for some magnitudes of data. The ifort compiler handles this by returning asterisks
+        # for all cases when outputting e.g. 'G14.0'. For this reasons we exit early when d=0
+        if d == 0:
+            return '*' * w
+
         nb = 0
         save_scale_factor = state['scale']
         exp_d = 10 ** d
         if (0.0 <= tmp < (0.1 - 0.05 / exp_d)) or \
-            (tmp >= (exp_d - 0.5)):
+                (tmp >= (exp_d - 0.5)):
             ftype = 'E'
         else:
             mag = int(abs(round(math.log10(tmp))))
-            low = lambda mag, d : 10 ** (mag - 1) - 5 * 10 ** (-d - 1 + mag) 
-            high = lambda mag, d : 10 ** mag - 0.5 * 10 ** (-d + mag)
+            def low(mag, d): return 10 ** (mag - 1) - 5 * 10 ** (-d - 1 + mag)
+            def high(mag, d): return 10 ** mag - 0.5 * 10 ** (-d + mag)
             while tmp < low(mag, d):
                 mag = mag - 1
             while tmp >= high(mag, d):
                 mag = mag + 1
             assert(low(mag, d) <= tmp < high(mag, d))
-            if e < 0:
+            if e is None:
                 nb = 4
             else:
                 nb = e + 2
@@ -311,7 +298,8 @@ def _compose_float_string(w, e, d, state, val, ftype):
                 d = d - mag
                 # d = -(mid - d - 1)
             state['scale'] = 0
-        out = _output_float(w, d, e, state, ftype, buff, sign_bit, zero_flag, ndigits, edigits)
+        out = _output_float(w, d, e, state, ftype, buff,
+                            sign_bit, zero_flag, ndigits, edigits)
         state['scale'] = save_scale_factor
         # TODO: this may not be right ...
         if nb > 0:
@@ -322,14 +310,14 @@ def _compose_float_string(w, e, d, state, val, ftype):
         if len(out) > (w + nb):
             out = '*' * (w + nb)
         return out
-        
+
 
 def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigits):
 
     # nbefore - number of digits before the decimal point
     # nzero - number of zeros after the decimal point
     # nafter - number of digits after the decimal point
-    # nzero_real - number of zeros after the decimal point regardles of the precision
+    # nzero_real - number of zeros after the decimal point regardless of the precision
 
     # Some hacks to change None to -1 (C convention)
     if w is None:
@@ -337,13 +325,11 @@ def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigit
     if e is None:
         e = -1
     nzero_real = -1
-    sign = _calculate_sign(state, sign_bit)
+
     # Some debug
     if d != 0:
         assert(buff[2] in ['.', ','])
         assert(buff[ndigits + 2] == 'e')
-    # Read in the exponent
-    ex = int(buff[ndigits + 3:]) + 1
     # Handle zero case
     if zero_flag:
         ex = 0
@@ -357,9 +343,14 @@ def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigit
         # This case does not include a decimal point
         if (w == 1) and (ft == 'F'):
             if state['incl_plus']:
-                return '*' # This is ifort behaviour
+                return '*'  # This is ifort behaviour
             else:
-                return '.' # CHANGED: Was '0'
+                return '.'  # CHANGED: Was '0'
+    else:
+        # Read in the exponent
+        ex = int(buff[ndigits + 3:]) + 1
+        sign = _calculate_sign(state, sign_bit)
+
     # Get rid of the decimal and the initial sign i.e. normalise the digits
     digits = buff[1] + buff[3:]
     # Find out where to place the decimal point
@@ -378,15 +369,17 @@ def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigit
         expchar = None
     elif ft in ['E', 'D']:
         i = state['scale']
-        if (d <= 0) and (i == 0):
-            raise InvalidFormat("Precision not greater than zero in format specifier 'E' or 'D'")
-        if (i <= -d) or (i >= (d + 2)):
-            raise InvalidFormat("Scale factor out of range in format specifier 'E' or 'D'")
+        if (d < 0) and (i == 0):
+            raise InvalidFormat(
+                "Precision less than zero in format specifier 'E' or 'D'")
+        if (i < -d) or (i >= (d + 2)):
+            raise InvalidFormat(
+                "Scale factor out of range in format specifier 'E' or 'D'")
         if not zero_flag:
             ex = ex - i
         if i < 0:
             nbefore = 0
-            nzero = -i;
+            nzero = -i
             nafter = d + i
         elif i > 0:
             nbefore = i
@@ -422,7 +415,8 @@ def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigit
     # Round the value
     if (nbefore + nafter) == 0:
         ndigits = 0
-        if (nzero_real == d) and (int(digits[0]) >= 5): # n.b. character comparison not very pythonic!
+        # n.b. character comparison not very pythonic!
+        if (nzero_real == d) and (int(digits[0]) >= 5):
             # We rounded to zero but shouldn't have
             nzero = nzero - 1
             nafter = 1
@@ -437,10 +431,10 @@ def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigit
             while i >= 0:
                 digit = int(digits[i])
                 if digit != 9:
-                    digits = _swapchar(digits, i, str(digit + 1))
+                    digits = _swapchar(digits, i, digit + 1)
                     break
                 else:
-                    digits = _swapchar(digits, i, '0')
+                    digits = _swapchar(digits, i, 0)
                 i = i - 1
             # Did the carry overflow?
             if i < 0:
@@ -482,18 +476,7 @@ def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigit
                 edigits = e + 2
     else:
         edigits = 0
-    # Zero values always output as positive, even if the value was egative before rounding
-    i = 0
-    while i < ndigits:
-        if digits[i] != '0':
-            break
-        i = i + 1
-    if i == ndigits:
-        # The output is zero so set sign accordingly
-        if PROC_SIGN_ZERO:
-            sign = _calculate_sign(state, sign_bit)
-        else:
-            sign = _calculate_sign(state, False)
+
     # Pick a field size if none was specified
     if w <= 0:
         w = nbefore + nzero + nafter + 1 + len(sign)
@@ -502,85 +485,78 @@ def _output_float(w, d, e, state, ft, buff, sign_bit, zero_flag, ndigits, edigit
     if sign != '':
         nblanks = nblanks - 1
     # TODO: Find out what this is
-    if G0_NO_BLANKS: # dtp->u.p.g0_no_blanks
+    if G0_NO_BLANKS:  # dtp->u.p.g0_no_blanks
         w = w - nblanks
         nblanks = 0
     # Check value fits in specified width
-    if (nblanks < 0) or (edigits == -1): 
+    if (nblanks < 0) or (edigits == -1):
         return '*' * w
     # See if we have space for a zero before the decimal point
-    if (nbefore == 0) and (nblanks > 0):
-        leadzero = True
+    leadzero = (nbefore == 0) and (nblanks > 0)
+    if leadzero:
         nblanks = nblanks - 1
-    else:
-        leadzero = False
-    out = ''
+
+    out = StringIO()
     # Pad to full field width
-    if (nblanks > 0) and not PROC_NO_LEADING_BLANK: # dtp->u.p.no_leading_blank
-        out = out + ' ' * nblanks
+    if (nblanks > 0) and not PROC_NO_LEADING_BLANK:  # dtp->u.p.no_leading_blank
+        out.write(' ' * nblanks)
     # Attach the sign
-    out = out + sign
+    out.write(sign)
     # Add the lead zero if necessary
     if leadzero:
-        out = out + '0'
+        out.write('0')
     # Output portion before the decimal point padded with zeros
     if nbefore > 0:
         if nbefore > ndigits:
-            out = out + digits[:ndigits] + (' ' * (nbefore - ndigits))
+            out.write(digits[:ndigits] + (' ' * (nbefore - ndigits)))
             digits = digits[ndigits:]
             ndigits = 0
         else:
             i = nbefore
-            out = out + digits[:i]
+            out.write(digits[:i])
             digits = digits[i:]
             ndigits = ndigits - i
+
     # Output the decimal point
-    out = out + PROC_DECIMAL_CHAR
+    out.write(PROC_DECIMAL_CHAR)
+
     # Output the leading zeros after the decimal point
     if nzero > 0:
-        out = out + ('0' * nzero)
+        out.write(('0' * nzero))
+
     # Output the digits after the decimal point, padded with zeros
     if nafter > 0:
-        if nafter > ndigits:
-            i = ndigits
-        else:
-            i = nafter
-        zeros = '0' * (nafter - i)
-        out = out + digits[:i] + zeros
-        digits = digits[nafter:]
-        ndigits = ndigits - nafter
+        i = min(nafter, ndigits)
+        out.write(digits[:i])
+        out.write('0' * (nafter - i))
+
     # Output the exponent
     if expchar is not None:
         if expchar != ' ':
-            out = out + expchar
+            out.write(expchar)
             edigits = edigits - 1
-        fmt = '%+0' + str(edigits) + 'd'
-        tmp_buff = fmt % ex
+        tmp_buff = f'{ex:+0{edigits}d}'
         # if not state['blanks_as_zeros']:
         if PROC_NO_LEADING_BLANK:
-            tmp_buf = tmp_buff + (nblanks * ' ')
-        out = out + tmp_buff
-    return out
+            tmp_buff = tmp_buff + (nblanks * ' ')
+        out.write(tmp_buff)
+    return out.getvalue()
 
 
 def _calculate_sign(state, negative_flag):
-    s = ''
     if negative_flag:
-        s = '-'
-    elif state['incl_plus']:
-        s = '+'
-    else:
-        s = ''
-    return s
+        return '-'
+    if state['incl_plus']:
+        return '+'
+    return ''
 
-def _swapchar(s, ind, newch):
+
+def _swapchar(s: str, ind: int, newch: int):
     '''
     Helper function to make chars in a string mutableish
     '''
-    if 0 < ind >= len(s):
-        raise IndexError('index out of range')
-    return s[:ind] + newch + s[ind+1:]
-    
+    return f"{s[:ind]}{newch}{s[ind+1:]}"
+
 
 def _compose_a_string(w, state, val):
     # f77 spec 13.5.11 covers a editing
@@ -611,7 +587,7 @@ def _compose_l_string(w, state, val):
 
 
 def _compose_i_string(w, m, state, val):
-    # f77 spec 13.5.9.1 covers integer editing 
+    # f77 spec 13.5.9.1 covers integer editing
     null_field = False
     # be pythonic in what values to accept, if it looks like an integer, then
     # so be it
@@ -638,6 +614,7 @@ def _compose_i_string(w, m, state, val):
         int_string = int_string.rjust(w)
     return int_string
 
+
 def _get_sign(val, incl_plus):
     if val >= 0:
         if incl_plus == True:
@@ -647,6 +624,7 @@ def _get_sign(val, incl_plus):
     else:
         return '-'
 
+
 def _compose_boz_string(w, m, state, val, ftype):
     try:
         val = int(val)
@@ -682,15 +660,18 @@ def _compose_boz_string(w, m, state, val, ftype):
         # Hex case
         # Fortran uses Two's complement to represent negative numbers, this
         # restricts hex values to sys.maxint, if greater than this overflow,
-        # however this is not perfoect as sys.maxint is Pythons build maximum
+        # however this is not perfect as sys.maxint is Pythons build maximum
         # rather than the systems bit size which is difficult to find reliably
         # across platforms
-        if abs(val) > config.PROC_MAXINT:
-            return '*' * w
-        if val < 0:
-            s = '%X' % ((config.PROC_MAXINT * 2) + 2 + val)
-        else:
+        if config.PROC_MAXINT is None:
             s = '%X' % val
+        else:
+            if abs(val) > config.PROC_MAXINT:
+                return '*' * w
+            if val < 0:
+                s = '%X' % (config.PROC_MAXINT * 2 + val)
+            else:
+                s = '%X' % val
     if m is None:
         s = s.rjust(w)
     else:
@@ -700,23 +681,23 @@ def _compose_boz_string(w, m, state, val, ftype):
     else:
         return s
 
-def _write_string(record, sub_string, pos):
+
+def _write_string(record: StringIO, sub_string: str, pos: int):
     '''Function that actually writes the generated strings to a 'stream'''''
-    new_pos = pos + len(sub_string)
     # pad if required with blanks - i.e. input after a tr edit descriptor - see
     # f77 format sec. 13.5.3
-    if pos > len(record):
-        record = record.ljust(pos)
-        out =  record + sub_string
-    elif pos == len(record):
-        out = record + sub_string
-    elif pos < len(record):
-        out = record[:pos] + sub_string + record[new_pos:]
-    return (new_pos, out)
+    record_len = record.tell()
+    if pos > record_len:
+        blanks = " " * (pos - record_len)
+        record.write(blanks)
+    elif pos < record_len:
+        record.seek(max(0, pos))
+
+    record.write(sub_string)
+    return (record.tell(), record)
 
 
 def left_pad(sub_string, width, pad_char):
     # Python 2.3 does not have the character argument to rjust
     padding = pad_char * (width - len(sub_string))
     return padding + sub_string
-
